/*
 * Copyright (c) 2012-2013 Stanford University
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR(S) DISCLAIM ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL AUTHORS BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdint.h>

#include <set>
#include <string>
#include <iostream>

#include <ori/udsrepo.h>

using namespace std;

extern UDSRepo repository;

int
cmd_branch(int argc, char * const argv[])
{
    strwstream req;
    string status;

    if (argc == 1) {
	string branch;

        req.writePStr("branch");
        req.writePStr("get");
        strstream resp = repository.callExt("FUSE", req.str());
        if (resp.ended()) {
            cout << "branch failed with an unknown error!" << endl;
            return 1;
        }

        resp.readLPStr(branch);

	cout << branch << endl;

	return 0;
    }

    if (argc == 2) {
        req.writePStr("branch");
        req.writePStr("set");
        req.writeLPStr(argv[1]);
        strstream resp = repository.callExt("FUSE", req.str());
        if (resp.ended()) {
            cout << "branch failed with an unknown error!" << endl;
            return 1;
        }

        resp.readLPStr(status);
        if (status != "") {
            cout << status << endl;
            return 1;
        }

        return 0;
    }

    cout << "Wrong number of arguments!" << endl;

    return 1;
}

