/*
 * Copyright (c) 2012 Stanford University
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR(S) DISCLAIM ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL AUTHORS BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdint.h>

#include <string>
#include <iostream>

#include <oriutil/debug.h>
#include <ori/localrepo.h>

using namespace std;

extern LocalRepo repository;

/*
 * Strip all meta-data including reference counts.
 */
int
cmd_stripmetadata(int argc, char * const argv[])
{
    if (argc != 1) {
        cout << "stripmetadata takes no arguments!" << endl;
        cout << "Usage: ori stripmetadata" << endl;
    }

    //if (!repository.stripMetadata())
    //    return 1;
    //return 0;

    // TODO
    printf("Strip metadata not supported!!!!");
    NOT_IMPLEMENTED(false);
    return 1;
}

