/*
 * Copyright (c) 2012 Stanford University
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR(S) DISCLAIM ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL AUTHORS BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __LOGGING_H__
#define __LOGGING_H__

void ori_fuse_log_enable();
void ori_fuse_log(const char *what, ...)
    __attribute__((format(printf, 1, 2)));

#if defined(DEBUG)
#define FUSE_LOG(fmt, ...) ori_fuse_log(fmt "\n", ##__VA_ARGS__)
#define FUSE_PLOG(fmt, ...) ori_fuse_log(fmt "\n", ##__VA_ARGS__)
#elif defined(ORI_PERF)
#define FUSE_LOG(fmt, ...)
#define FUSE_PLOG(fmt, ...) ori_fuse_log(fmt "\n", ##__VA_ARGS__)
#else
#define FUSE_LOG(fmt, ...)
#define FUSE_PLOG(fmt, ...)
#endif

#endif /* __LOGGING_H__ */

