/*
 * Copyright (c) 2012 Stanford University
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR(S) DISCLAIM ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL AUTHORS BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __HTTPDEFS_H__
#define __HTTPDEFS_H__

// HTTP Paths
#define ORIHTTP_PATH_ID         "/id"
#define ORIHTTP_PATH_VERSION    "/version"
#define ORIHTTP_PATH_HEAD       "/HEAD"
#define ORIHTTP_PATH_INDEX      "/index"
#define ORIHTTP_PATH_COMMITS    "/commits"
#define ORIHTTP_PATH_CONTAINS   "/contains"
#define ORIHTTP_PATH_GETOBJS    "/getobjs"
#define ORIHTTP_PATH_OBJINFO    "/objinfo/"

#endif /* __HTTPDEFS_H__ */

